;; Generated package description from listen.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "listen" "0.10" "Audio/Music player" '((emacs "29.1") (persist "0.6") (taxy "0.10") (taxy-magit-section "0.13") (transient "0.5.3")) :commit "e0a243d3c6ba0a8a14d1d16168ac23441bd5e6c9" :authors '(("Adam Porter" . "adam@alphapapa.net")) :maintainer '("Adam Porter" . "adam@alphapapa.net") :keywords '("multimedia") :url "https://github.com/alphapapa/listen.el")
